package org.simantics.databoard.binding.reflection;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Arrays;

import org.simantics.databoard.annotations.ArgumentImpl;
import org.simantics.databoard.annotations.Arguments;

public class BindingRequest {
	/** Requested class */
    public final Class<?> clazz; 
    /** Annotations */
    public final Annotation[] annotations;
    
    public final Annotation[] NO_ANNOTATIONS = new Annotation[0];
    
    transient int hash;
    
    public BindingRequest(Class<?> clazz, Annotation...annotations)
    {
        assert annotations!=null;
        this.clazz = clazz;
        Annotation[] classAnnotations = clazz.getAnnotations();
        if (classAnnotations!=null && classAnnotations.length>0) {
            this.annotations = new Annotation[classAnnotations.length + annotations.length];
            System.arraycopy(annotations, 0, this.annotations, 0, annotations.length);
            System.arraycopy(classAnnotations, 0, this.annotations, annotations.length, classAnnotations.length);
        } else {
        	this.annotations = annotations;
       	}
        hash = clazz.hashCode();
        for (Annotation a : annotations)
            hash = 7*hash + a.hashCode();            
    }
    
    @SuppressWarnings("unchecked")
    public boolean hasAnnotation(Class<?> annotationClass) 
    {
        for (Annotation a : annotations)
            if (annotationClass.equals(a.annotationType())) return true;
        return false;
    }
    
    @SuppressWarnings("unchecked")
    public <A extends Annotation> A getAnnotation(Class<A> annotationClass) 
    {
        for (Annotation a : annotations)
        {
            if (annotationClass.equals(a.annotationType()))
                return (A) a;
        }
        return null;
    }
    @Override
    public int hashCode() {
        return hash;
    }
    @Override
    public boolean equals(Object obj) {
        if (obj==null) return false;
        if (obj instanceof BindingRequest==false) return false;
        BindingRequest other = (BindingRequest) obj;
        return other.clazz.equals(clazz) &&
            Arrays.deepEquals(annotations, other.annotations);
    }
    
    /**
     * Return a version of annotations list, where given set of annotations and
     * a number of class arguments were dropped. 
     * 
     * @param argumentsToDrop the number of class arguments to drop
     * @param annotationsToDrop annotation to drop
     * @return request without argument annotation
     */
    public Annotation[] dropAnnotations(int argumentsToDrop, Annotation...annotationsToDrop)
    {
    	ArrayList<Annotation> result = new ArrayList<Annotation>( annotations.length );
    	nextA:
    	for (Annotation a : annotations) {
    		for (Annotation b : annotationsToDrop) 
    			if (a==b) continue nextA;
    		if (a instanceof Arguments && argumentsToDrop>0) {
    			Arguments c = ArgumentImpl.dropArguments((Arguments) a, argumentsToDrop);
    			if (c!=null) result.add(c);
    		} else result.add(a);
    	}
    	Annotation[] newAnnotations = result.toArray( new Annotation[result.size()] );
    	return newAnnotations;
    }
    
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(clazz.getName());
        if ( annotations!=null && annotations.length>0 ) {
            sb.append('(');
            for (int i=0; i<annotations.length; i++) {
                Annotation a = annotations[i];
                if (i>0) sb.append(", ");
                sb.append(a);
            }           
            sb.append(')');
        }
        
        return sb.toString();
    }

}
