/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.history;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.Datatypes;
import org.simantics.databoard.Historian;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.ArrayAccessor;
import org.simantics.databoard.accessor.MapAccessor;
import org.simantics.databoard.accessor.RecordAccessor;
import org.simantics.databoard.accessor.VariantAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.OptionalValueAssigned;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.interestset.MapInterestSet;
import org.simantics.databoard.accessor.reference.ComponentReference;
import org.simantics.databoard.accessor.reference.NameReference;
import org.simantics.databoard.accessor.reference.IndexReference;
import org.simantics.databoard.accessor.reference.KeyReference;
import org.simantics.databoard.accessor.reference.ChildReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.NumberBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.mutable.MutableVariant;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.Datatype;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.NumberType;
import org.simantics.databoard.type.RecordType;

/**
 * Sample collector reads values from a datasource, samples them and 
 * adds to a RecordingSession.
 * 
 * Recording Session is a history storage format. 
 * http://dev.simantics.org/index.php/Databoard_Specification#History_Contract
 * 
 * There is a contract for the structure of data sources.
 * http://dev.simantics.org/index.php/Databoard_Specification#Datasource_Contract
 * 
 * Sample Collection Steps:
 *  
 *   1. Initialize Datasource and have it in paused state
 *   2. Initialize Sample Collector with Subscription Parameters
 *      a. Verify subsribed variables
 *      b. Initialize variables in history, add variable specific Recording-structures
 *      c. Read initial values from Datasource
 *   3. Start Sampling
 *      a. Initialize new Segment to each Recording
 *      b. Write initial values to History
 *      c. Add listener to Datasource
 *      d. Start Datasource
 *      e. Convert incoming events them into samples
 *   4. Stop Sampling
 *      a. Stop Datasource
 *      b. Remove listener from Datasource
 *      c. Finalize each Segment. Write last known values
 * 
 * todo how to deliver errors during sampling to the user, log them? collect them?
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class SampleCollector {
	
	/** Current sampling configuration */
	SamplingConfiguration configuration;
	
	/** Accessor to Datasource */
	Accessor datasource;
	
	/** Accessor to history */
	Accessor recordingSession;
	
	/** Accessor to the recordings Map(nodeId, Recording) */ 
	MapAccessor recordings;

	/** */
	Object samplingConfiguration;
	
	/** Sampling is activated with start() */
	boolean active = false;
	
	/** NodeId, Variable map */
	Map<Variant, Variable> vars = new HashMap<Variant, Variable>();
 	
	/** Time variable, optional */
	Variable timeVariable;
	
	/** Step is a virtual time code, incremented on each event from datasoource */
	int step = 0;

	/** Binding of time variable */
	Binding timeBinding;
	
	/** Set to true in init if time is number type */
	boolean timeIsNumberType = false;

	/** Binding of time */
	MutableVariant currentTime = new MutableVariant();
	
	/**
	 * Create sample collector 
	 * @param datasource
	 * @param recordingSession
	 * @throws AccessorConstructionException 
	 */
	public SampleCollector(Accessor datasource, Accessor recordingSession) throws AccessorConstructionException 
	{
		this.datasource = datasource;
		this.recordingSession = recordingSession;
		this.recordings = recordingSession.getComponent( new NameReference("recordings") );
	}
	
	class Variable {
		/** Last Written Value */
		MutableVariant writtenValue = new MutableVariant();
		/** 
		 * Time of the written value in the binding of the time variable. 
		 * This field is not captured if interval is not used. 
		 */
		int writtenValueStep = -1;
		
		/** The time when values was written */
		Object writtenValueTime = null;
		
		/** Last known value */
		MutableVariant knownValue = new MutableVariant();
		/** 
		 * Time of the known value in the type and units of the time variable. 
		 * This field is not captured if interval is not used. 
		 */
		Object knownValueTime = null;
		/** The step when known value was assigned */
		int knownValueStep = -1;
		
		/** If true, this variable is written to history */
		boolean subscribed = true;
		
		/** Deadband in the binding of the variable or <code>null</code> */
		Number deadband;
		/** Interval in the binding of time variable or <code>null</code> */
		Number interval;		
		
		/** Node Id */
		Variant nodeId;
		
		/** Variable Info */
		Datatype type;
		/** Default binding */
		Binding binding;
		
		/** If true this is dense sampling segment, else this is sparse */
		boolean dense;
		
		/** Accessor to the primitive value in Datasource */
		Accessor valueAccessor;
		
		/** Accessor to the node in Datasource */
		RecordAccessor nodeAccessor;
		
		/** Accessor to recording in History */
		RecordAccessor recordingAccessor;

		/** Accessor to active segment, set in start() */
		ArrayAccessor segmentsAccessor;
		
		/** Accessor to active segment, set in start() */
		MapAccessor segmentAccessor;
	}
	
	/**
	 * Initialize sample collector with the given subscription parameters.
	 * This step verifies the subscribed variables, reads initial values and
	 * adds variable specific Recording-structures to the history. <p> 
	 *  
	 * The datasource must be initialized but paused. Variable initial values
	 * must be available.<p> 
	 * 
	 * @param configuration subscription confifguration
	 */
	public void init(SamplingConfiguration configuration) 
	throws SamplingException 
	{
		try {
			if (active) throw new SamplingException("Collector is already active");
			
			timeIsNumberType = false;
			step = 0;

			// Get accessor to datasource's nodes
			ChildReference node_to_value_reference = ChildReference.parsePath("value/o/v");			
			Accessor nodes = datasource.getComponent(ChildReference.parsePath("nodes"));
			
			// Initialize Variable, accessors, type, binding, and verify the variables exists
			int count = configuration.subscriptions.length;
			vars.clear();
			for (int i=0; i<count; i++) {
				Variable var = new Variable();
				Subscription s = configuration.subscriptions[i];
				Variant nodeId = s.nodeId;
				vars.put(nodeId, var);
				
				// This is time variable
				if ( configuration.timeId.equals(nodeId) ) {
					timeVariable = var;
				}
				
				var.nodeId = nodeId;
				var.nodeAccessor = nodes.getComponent( new KeyReference(nodeId) );
				var.valueAccessor = var.nodeAccessor.getComponent( node_to_value_reference );
				var.type = var.valueAccessor.type();
				var.binding = Bindings.getMutableBinding( var.type );				
			}
			
			// Time variable was not subscribed, we still need it as variable
			if (timeVariable == null && configuration.timeId!=null) {
				Variable var = new Variable();
				Variant nodeId = configuration.timeId;
				vars.put(nodeId, var);
				timeVariable = var;
				
				var.subscribed = false;
				var.nodeId = nodeId;
				var.nodeAccessor = nodes.getComponent( new KeyReference(nodeId) );
				var.valueAccessor = var.nodeAccessor.getComponent( node_to_value_reference );
				var.type = var.valueAccessor.type();
				var.binding = Bindings.getMutableBinding( var.type );				
			}
			
			if (timeVariable != null) {
				timeIsNumberType = timeVariable.type instanceof NumberType;
				timeBinding = timeVariable.binding;
			} else {
				timeIsNumberType = true;
				timeBinding = Bindings.INTEGER;
			}
			
			// Initialize Intervals and Deadbands
			for (int i=0; i<count; i++) {
				Subscription s = configuration.subscriptions[i];
				Variable var = vars.get(s.nodeId);
				var.knownValueStep = 0;
				
				// Has interval
				if (s.interval != null) {
					if (timeVariable == null) {
						throw new SamplingException("Cannot subscribe "+s.nodeId+" with interval, because there is no Time variable.");						
					}
					
					if (!timeIsNumberType) {
						throw new SamplingException("Cannot subscribe "+s.nodeId+" with interval, because the Time variable is not a number type. Time Variable is "+timeVariable.type);
					}

					if (s.interval.getBinding() instanceof NumberBinding == false) { 
						throw new SamplingException("Cannot subscribe "+s.nodeId+", becasuse the interval is a non-number.");
					}
					
					// Convert interval to the format of the time variable					
					var.interval = (Number) Bindings.adapt(s.interval.getValue(), s.interval.getBinding(), timeVariable.binding);
				}

				// Deadband
				if (s.deadband != null) {
					if (var.binding instanceof NumberBinding == false) { 
						throw new SamplingException("Cannot subscribe "+s.nodeId+" with deadband, because the variable is a non-number. The variable is "+var.type);
					}

					if (s.deadband.getBinding() instanceof NumberBinding == false) { 
						throw new SamplingException("Cannot subscribe "+s.nodeId+", because the deadband is a non-number. Deadband is "+s.deadband.type());
					}
					
					// Convert deadband to the format of the variable
					var.deadband = (Number) Bindings.adapt(s.deadband.getValue(), s.deadband.getBinding(), var.binding);					
				}
				
			}
			
			// Initialize history
			for (Variable var : vars.values()) {				
				if (!var.subscribed) continue;
				
				// Get-or-Create recording
				if (recordings.containsKey(Bindings.VARIANT, var.nodeId)) {
					// Use existing recording
					var.recordingAccessor = ((VariantAccessor) recordings.getValueAccessor(Bindings.VARIANT, var.nodeId)).getContentAccessor();
					
					// Assert time and value types are correct
					MapType recordingType = (MapType) ((ArrayType)((RecordType) var.recordingAccessor.type()).getComponentType("segments")).componentType;
					Datatype recordingTimeType = recordingType.getKeyType(); 
					Datatype recordingValueType = recordingType.getValueType(); 
					if (!recordingTimeType.equals( timeBinding.type() )) {
						throw new SamplingException("Datasource + History time type mismatch; recording for "+var.nodeId+" has "+recordingTimeType+", and datasource "+timeBinding.type());
					}
					if (!recordingValueType.equals( var.type )) {
						throw new SamplingException("Datasource + History value type mismatch; recording for "+var.nodeId+" has "+recordingValueType+", and datasource "+var.type);
					}
					
				} else {
					// Create recording
					Datatype recordingType = Historian.createRecordingType( timeBinding.type(), var.type );
					RecordBinding binding = (RecordBinding) Bindings.getMutableBinding( recordingType );
					Object instance = binding.createDefault();
					
					// id
					binding.setComponent(instance, 0, var.nodeId);
					
					// Labels
					Datatype localized_text_type = Datatypes.getDatatype("LocalizedText");
					Binding localized_text_binding = Bindings.getMutableBinding(localized_text_type);
					Object labels = var.nodeAccessor.getFieldAccessor("labels").getValue( localized_text_binding );
					binding.setComponent(instance, 1, labels);
					
					// Add recording to RecordingSession
					recordings.put(Bindings.VARIANT, var.nodeId, Bindings.VARIANT, new Variant(binding, instance) );
					// Get accessor to Recording 
					var.recordingAccessor = ((VariantAccessor) recordings.getValueAccessor(Bindings.VARIANT, var.nodeId)).getContentAccessor();
				}

			}
			
			
		} catch (AccessorConstructionException e) {
			throw new SamplingException(e);
		} catch (BindingException e) {
			throw new SamplingException(e);
		} catch (AccessorException e) {
			throw new SamplingException(e);
		} catch (AdaptException e) {
			throw new SamplingException(e);
		}
	}
	
	/**
	 * Start sampling process. This should be called just before the 
	 * datasource is started. In this step, Segment-structure is initialized 
	 * for each Recording. Listener is attached to the datasource, which 
	 * converts invoming events into samples. 
	 * 
	 * @throws SamplingException
	 */
	public void start() throws SamplingException {
		try {
			// Create segments for each variable
			for (Variable var : vars.values()) {	
				if (!var.subscribed) continue;
				var.segmentsAccessor = var.recordingAccessor.getFieldAccessor("segments");
				MapType segmentType = (MapType) var.segmentsAccessor.type().componentType;
				MapBinding segmentBinding = Bindings.getMutableBinding( segmentType );
				
				// Add empty segment
				var.segmentsAccessor.add(segmentBinding, segmentBinding.create());
				
				// Get accessor to the segment in the history
				var.segmentAccessor = var.segmentsAccessor.getAccessor( var.segmentsAccessor.size()-1 );
			}
		
			// Calculate pseudo time; last time stamp + 1 
			if (timeVariable == null) {				
				// Get the last known timecode and +1
				step = 0;
				for (Variable var : vars.values()) {
					ArrayAccessor segments = var.recordingAccessor.getFieldAccessor("segments");
					int count2 = segments.size();
					for (int i = 0; i<count2; i++) {
						MapAccessor segment = segments.getAccessor(i);
						Integer lastKey = (Integer) segment.getLastKey( Bindings.INTEGER );
						if (lastKey!=null) {
							step = Math.max(lastKey+1, step);
						}
					}
				}
			}
		
			// Write initial values to history
			// Get initial values
			for (Variable var : vars.values()) {
				Object ov = var.valueAccessor.getValue( var.binding );
				var.knownValue.setValue( var.binding, ov );				
				var.knownValueStep = step;
			}
			
			for (Variable var : vars.values()) {
				Object ov = var.valueAccessor.getValue( var.binding );
				var.knownValue.setValue( var.binding, ov );				
				var.knownValueStep = step;
			}
			

			// Acquire time code
			Object timeValue;
			if (timeVariable == null) {
				timeValue = step;
			} else {
				timeValue = timeVariable.knownValue.getValue(timeBinding);						
			}			
			
			for (Variable var : vars.values()) {
				var.knownValueTime = timeValue;		
			}

			// Write initial values
			for (Variable var : vars.values()) {				
				if (!var.subscribed) continue;
				var.segmentAccessor.put( timeBinding, timeValue, var.knownValue.getBinding(), var.knownValue.getValue() );
				var.writtenValue.setValue( var.knownValue );
				var.writtenValueStep = step;
				var.writtenValueTime = timeValue;
			}
			step++;
		
			// Create a listener InterestSet that listens to all variables 
			// Create interest set
			InterestSet nodeInterest = InterestSet.newInterestSet( Datatypes.getDatatype("Node"), true, true, true); 
			MapInterestSet is = new MapInterestSet();
			for (Variant nodeId : vars.keySet())
				is.addCompositeInterest(nodeId, nodeInterest);
			
			// Add listener to Datasource
			Accessor nodesAccessor = datasource.getComponent( new NameReference("nodes") );
			nodesAccessor.addListener(listener, is, null);
		} catch (AccessorConstructionException e) {
			throw new SamplingException( e );
		} catch (AccessorException e) {
			throw new SamplingException( e );
		} catch (BindingException e) {
			throw new SamplingException( e );
		} catch (AdaptException e) {
			throw new SamplingException( e );
		}
		
		active = true;
			
	}
	
	/**
	 * Stop collecting samples. Writes down last remaining values.
	 */
	public synchronized void stop() throws SamplingException {
		active = false;
		try {
			Accessor nodesAccessor = datasource.getComponent( new NameReference("nodes") );
			nodesAccessor.removeListener( listener );
		} catch (AccessorConstructionException e) {
			throw new SamplingException( e );
		} catch (AccessorException e) {
			throw new SamplingException( e );
		}
		
		// Write final values to history
		try {
			Object timeValue;
			if (timeVariable == null) {
				timeValue = step;
			} else {
				timeValue = timeVariable.knownValue.getValue(timeBinding);						
			}
			
			for (Variable var : vars.values()) {
				if (!var.subscribed) continue;				
				if (var.knownValueTime.equals( var.writtenValueStep )) continue;
				if (timeVariable==null) {
					var.segmentAccessor.put(timeBinding, timeValue, var.knownValue.getBinding(), var.knownValue.getValue());
				} else {
					var.segmentAccessor.put(timeBinding, timeValue, var.knownValue.getBinding(), var.knownValue.getValue());
				}
				var.writtenValue.setValue( timeBinding, timeValue );
			}
			step++;
		} catch (AccessorException e) {
			throw new SamplingException( e );
		} catch (AdaptException e) {
			throw new SamplingException( e );
		}
		
	}	

	/**
	 * A listener attached to an "nodes"-map of datasource.  
	 */
	Accessor.Listener listener = new Accessor.Listener() {
		@Override
		public void onEvents(Collection<Event> events) {
			synchronized(SampleCollector.this) {
				if (!active) return;
				
				int newStep = step + 1;
				MutableVariant value = new MutableVariant();
				MutableVariant nodeId = new MutableVariant();
				// Parse events and get time value 
				for (Event e : events) {
					try {
						if (!readEvent(e, value, nodeId)) continue;
						Variable var = vars.get( nodeId.getValue( Bindings.VARIANT ) );
						if (var == null) continue;
					
						var.knownValue.setValue(value);						
						var.knownValueStep = newStep;
					} catch (BindingException be) {
						be.printStackTrace();
					} catch (AdaptException ae) {
						ae.printStackTrace();
					}		
				}
				
				// Acquire time code
				Object timeValue;
				if (timeVariable == null) {
					timeValue = newStep;
				} else {
					try {
						timeValue = timeVariable.knownValue.getValue(timeBinding);						
					} catch (AdaptException e1) {
						e1.printStackTrace();
						return;
					}
				}
				
				// Time code changed, update 
				if (timeVariable.knownValueStep == newStep) {
					for (Variable v : vars.values()) {
						try {
							// Value changed on this step
							boolean valueChangedOnThisStep = v.knownValueStep == newStep;
							// Value has changed, but not written
							boolean valueChanged = v.knownValueStep != v.writtenValueStep; 
							
							// Update knownValueField
							if (valueChangedOnThisStep) v.knownValueTime = timeValue;						

							if (!v.subscribed) continue;
							
							// Does time-axis allow to write value
							boolean timeToWrite = true;
							if (v.interval != null) {
								// Calculate elapsed time
								double lastWrittenTime = v.writtenValueTime != null ? ((Number)v.writtenValueTime).doubleValue() : 0;
								double currentTime = ((NumberBinding)timeBinding).getValue(timeValue).doubleValue();
								double elapsedTime = currentTime - lastWrittenTime;
								timeToWrite = elapsedTime > v.interval.doubleValue(); 
							}
							
							boolean changeEnoughToWrite = valueChanged;
							if (valueChanged && v.deadband != null) {
								double lastWrittenValue = v.writtenValue != null ? ((Number)v.writtenValue.getValue()).doubleValue() : 0;
								double currentValue = (Double) v.knownValue.getValue( Bindings.DOUBLE );
								double diff = Math.abs( lastWrittenValue - currentValue );
								changeEnoughToWrite = diff > v.deadband.doubleValue(); 
							}
							
							// Write the value
							if (timeToWrite && changeEnoughToWrite) {
								v.segmentAccessor.put(timeBinding, timeValue, v.knownValue.getBinding(), v.knownValue.getValue());
								v.writtenValue.setValue( v.knownValue );
								v.writtenValueStep = step;
								v.writtenValueTime = timeValue;							
							}
							
						} catch (AdaptException e) {
							e.printStackTrace();
						} catch (AccessorException e) {
							e.printStackTrace();
						} catch (BindingException e) {
							e.printStackTrace();
						}
					}
				}				
					
				step++;				
			}
		}
	};
	
	/**
	 * Read an Event send by nodes Map and parse it into a new value
	 * 
	 * @param e
	 * @param result variant where value is written
	 * @param nodeId value to be set with node id
	 * @return true if new value was read
	 * @throws BindingException 
	 */
	boolean readEvent(Event e, MutableVariant result, MutableVariant nodeId) throws BindingException {
		if (e.reference==null || e.reference instanceof KeyReference==false) return false;
		KeyReference kr = (KeyReference) e.reference;
		nodeId.setValue( kr.key.getBinding(), kr.key.getValue() );
//		Variable var = vars.get(kr.key);
//		if (var == null) return false;
		
		// Event applies to the whole Node-record, read only the value
		if (e instanceof ValueAssigned && kr.getChildReference()==null) {
			ValueAssigned va = (ValueAssigned) e;
			RecordBinding rb = (RecordBinding) va.newValue.getBinding();
			Object ov = rb.getComponent(va.newValue, 3);
			OptionalBinding ob = (OptionalBinding) rb.getComponentBinding(3);
			Object vv = ob.getValue(ov);
			VariantBinding vb = (VariantBinding) ob.getComponentBinding();
			Object pv = vb.getContent(vv);
			result.setValue( vb.getContentBinding(vv), vb.getContent(vv) );
			return true;
		}
		
		// Event applies to "value" field
		ChildReference fr = kr.getChildReference();				
		if ( !( fr instanceof IndexReference && ((IndexReference)fr).getIndex()==3 ) &&
			 !( fr instanceof NameReference  && ((NameReference)fr).getName().equals("value") ) ) return false;

		// Event applies to the whole Optional
		if ( fr.getChildReference()==null && e instanceof ValueAssigned ) {
			ValueAssigned va = (ValueAssigned) e;
			Object ov = va.newValue.getValue();
			OptionalBinding ob = (OptionalBinding) va.newValue.getBinding();
			Object vv = ob.getValue(ov);
			VariantBinding vb = (VariantBinding) ob.getComponentBinding();
			Object pv = vb.getContent(vv);
			result.setValue(vb, pv);
			return true;
		}
						
		// Event applies to value of Optional 
		if ( fr.getChildReference()==null && e instanceof OptionalValueAssigned ) {
			OptionalValueAssigned va = (OptionalValueAssigned) e;
			
			Object vv = va.newValue.getValue();
			VariantBinding vb = (VariantBinding) va.newValue.getBinding();
			Object pv = vb.getContent(vv);
			result.setValue(vb, pv);
			return true;
		}

		if (e instanceof ValueAssigned == false) return false;
		ValueAssigned va = (ValueAssigned) e;

		// Optional's Reference
		ChildReference or = fr.getChildReference();
		if (or == null || or instanceof ComponentReference == false) return false;
		
		ChildReference vr = or.getChildReference();
		// Event applies to Variant				
		if (vr == null) {
			Object vv = va.newValue.getValue();
			VariantBinding vb = (VariantBinding) va.newValue.getBinding();
			Object pv = vb.getContent(vv);
			result.setValue(vb, pv);			
			return true;
		}
		
		// Variant's Reference
		// Event applies to Value
		if (vr.getChildReference() == null) {
			result.setValue(va.newValue);
			return true;
		}
		
		// It is an event inside the value, eg. to a field of a record
		// WTF to do now?
		System.err.println("Don't know how to apply this event: "+e);
		return false;
	}
	
}

