/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.type;

import java.util.Set;

import org.simantics.databoard.accessor.reference.ChildReference;
import org.simantics.databoard.accessor.reference.IndexReference;
import org.simantics.databoard.accessor.reference.KeyReference;
import org.simantics.databoard.accessor.reference.LabelReference;
import org.simantics.databoard.accessor.reference.NameReference;
import org.simantics.databoard.annotations.Optional;
import org.simantics.databoard.util.IdentityPair;
import org.simantics.databoard.util.Limit;
import org.simantics.databoard.util.ObjectUtils;
import org.simantics.databoard.util.Range;

public class ArrayType extends Datatype {
    
    public Datatype componentType;
    
    @Optional Range length;
    
    public ArrayType() {}
    
    public ArrayType(Datatype componentType) {
    	this.componentType = componentType;
    }
    
    public ArrayType(Datatype componentType, String length) {
    	this.componentType = componentType;
    	setLength(length);
    }

    public ArrayType(Datatype componentType, Range length) {
    	this.componentType = componentType;
    	setLength(length);
    }
    
    @Override
    public int count() {
    	return 1;
    }
    
    @Override
    public Datatype getComponentType(int index) {
    	if (index!=0) throw new IllegalArgumentException();
    	return componentType;
    }
    
    @Override
    public Datatype getComponentType(ChildReference path) {
    	if (path==null) return this;
    	if (path instanceof KeyReference) throw new IllegalArgumentException("KeyReference is not supported in ArrayType"); 
    	if (path instanceof NameReference) throw new IllegalArgumentException("NameReference is not supported in ArrayType"); 
    	if (path instanceof IndexReference && ((IndexReference) path).index!=0) throw new IllegalArgumentException("Index out of bounds");
    	if (path instanceof LabelReference && !((LabelReference) path).label.equals("v")) throw new IllegalArgumentException("Unknown label");
    	return componentType.getComponentType(path.childReference);
    }
    
	@Override
	protected void collectSubtypes(Set<Datatype> subtypes, Set<Datatype> recursiveSubtypes) {
		componentType.collectSubtypes(subtypes, recursiveSubtypes);
	}
	
	@Override
	protected boolean deepEquals(Object obj,
			Set<IdentityPair<Datatype, Datatype>> compareHistory) {
		if (this==obj) return true;
		if (obj instanceof ArrayType == false) return false;
		ArrayType other = (ArrayType) obj;
		if (!ObjectUtils.objectEquals(other.length, length)) return false;
		
		return componentType.deepEquals(other.componentType, compareHistory);
	}
	
	@Override
	public int hashCode() {
		if (componentType==this) return 0;
		return 0x234ae +
			ObjectUtils.hashCode(componentType) * 13 +
			ObjectUtils.hashCode(length) * 17;
	}	

	@Override
	public void accept(Visitor1 v, Object obj) {
	    v.visit(this, obj);        
	}

	@Override
	public <T> T accept(Visitor<T> v) {
	    return v.visit(this);
	}

	public Datatype getComponentType() {
		return componentType;
	}

	public void setComponentType(Datatype componentType) {
		this.componentType = componentType;
	}

    public int minLength() {
    	if (length==null) return 0;
    	Limit l = length.getLower();
    	int value = l.getValue().intValue();
    	if (l.isExclusive()) value++;
    	return value;
    }
    
    public int maxLength() {
    	if (length==null) return Integer.MAX_VALUE;
    	Limit l = length.getUpper();
    	int value = l.getValue().intValue();
    	if (l.isExclusive()) value--;
    	return value;
    }
	
	public Range getLength() {
		return length;
	}

	public String getLengthStr() {
		return length==null ? null : length.toString();
	}
	
	public void setLength(String length) {
    	this.length = length == null ? null : Range.valueOfUnchecked(length);
	}
	
	public void setLength(Range range) {
		this.length = range;
	}
	
}
